/* 
*	Copyright (C) Chia-chen Kuo - April 2001
*
*  This file is part of DVD2AVI, a free MPEG-2 decoder
*	
*  DVD2AVI is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2, or (at your option)
*  any later version.
*   
*  DVD2AVI is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*   
*  You should have received a copy of the GNU General Public License
*  along with GNU Make; see the file COPYING.  If not, write to
*  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*
*  AAC demux portion comes from TSDemux 0.18 by Chia-chen Kuo
*
*/

#include "global.h"
#include "getbit.h"
#include "AC3Dec\ac3.h"
#include <stdio.h>
#include <tchar.h>

void dprintf(TCHAR *format, ...)
{
    va_list vl;
    TCHAR debugbuf[2048];

    va_start(vl, format);
    _vstprintf(debugbuf, format, vl);
    va_end(vl);

    OutputDebugString(debugbuf);
}

// Fill up read buffer? I guess.
#define LOCATE									\
    while (Rdptr >= (Rdbfr + BUFFER_SIZE)) {					\
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);			\
	if (Read < BUFFER_SIZE)							\
    	    Next_File();							\
	Rdptr -= BUFFER_SIZE;							\
    }

#define DECODE_AC3																\
{																				\
    size = 0;																	\
    while (Packet_Length > 0)													\
	{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
		{																		\
		size = ac3_decode_data(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, size);		\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
		if (Read < BUFFER_SIZE) Next_File();								\
		Rdptr = Rdbfr;														\
		}																		\
		else																	\
		{																		\
		size = ac3_decode_data(Rdptr, Packet_Length, size);					\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
		}																		\
	}																			\
}

#define DEMUX_AC3															\
    while (Packet_Length > 0)													\
{																			\
    if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
	fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, ac3[This_Track].file);	\
	Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
	if (Read < BUFFER_SIZE) Next_File();								\
	Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
	fwrite(Rdptr, Packet_Length, 1, ac3[This_Track].file);				\
	Rdptr += Packet_Length;												\
	Packet_Length = 0;													\
	}																		\
}

#define DEMUX_PCM														\
{																		\
    unsigned char tmp;													\
    size = 0;															\
    while (Packet_Length > 0)											\
	{																	\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)					\
		{																\
		memcpy(PCM_Buffer+size, Rdptr, BUFFER_SIZE+Rdbfr-Rdptr);	\
		size += BUFFER_SIZE+Rdbfr-Rdptr;							\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;					\
		Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);		\
		if (Read < BUFFER_SIZE) Next_File();						\
		Rdptr = Rdbfr;												\
		}																\
		else															\
		{																\
		memcpy(PCM_Buffer+size, Rdptr, Packet_Length);				\
		Rdptr += Packet_Length;										\
		size += Packet_Length;										\
		Packet_Length = 0;											\
		}																\
	}																	\
	for (i=0; i<size; i+=2)												\
	{																	\
	tmp = PCM_Buffer[i];											\
	PCM_Buffer[i] = PCM_Buffer[i+1];								\
	PCM_Buffer[i+1] = tmp;											\
	}																	\
}

#define DEMUX_MPA															\
    while (Packet_Length > 0)													\
{																			\
    if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
	fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, mpa[MPA_Track].file);		\
	Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);				\
	if (Read < BUFFER_SIZE) Next_File();								\
	Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
	fwrite(Rdptr, Packet_Length, 1, mpa[MPA_Track].file);				\
	Rdptr += Packet_Length;												\
	Packet_Length = 0;													\
	}																		\
}

#define DEMUX_DTS								\
    while (Packet_Length > 0) {							\
    if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)				\
	{									\
	fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, dts[This_Track].file);	\
	Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;				\
	Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);			\
	if (Read < BUFFER_SIZE) Next_File();					\
	Rdptr = Rdbfr;								\
	}									\
	else									\
	{									\
	fwrite(Rdptr, Packet_Length, 1, dts[This_Track].file);			\
	Rdptr += Packet_Length;							\
	Packet_Length = 0;							\
	}									\
    }

#define DEMUX_AAC()								\
    while (Packet_Length > 0) {							\
	if (Packet_Length + Rdptr > BUFFER_SIZE + Rdbfr) {			\
	    fwrite(Rdptr, BUFFER_SIZE + Rdbfr - Rdptr, 1, aac[AAC_Track].file);	\
	    Packet_Length -= BUFFER_SIZE + Rdbfr - Rdptr;			\
	    Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);		\
	    if (Read < BUFFER_SIZE)						\
		Next_File();							\
	    Rdptr = Rdbfr;							\
	} else {								\
	    fwrite(Rdptr, Packet_Length, 1, aac[AAC_Track].file);		\
	    Rdptr += Packet_Length;						\
	    Packet_Length = 0;							\
	}									\
    }

// skips ahead by "Packet_Length" bytes
#define SKIP_TRANSPORT_PACKET_BYTES(bytes_to_skip)				\
{										\
    int temp = (bytes_to_skip);							\
    while (temp > 0) { 								\
	if (temp + Rdptr > BUFFER_SIZE + Rdbfr) {				\
	    temp  -= BUFFER_SIZE + Rdbfr - Rdptr;				\
	    Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);		\
	    if (Read < BUFFER_SIZE)						\
		Next_File();							\
	    Rdptr = Rdbfr;							\
	} else { 								\
	    Rdptr += temp;							\
	    temp  = 0;								\
	} 									\
    }										\
    Packet_Length -= (bytes_to_skip);						\
}

// structure describing transport packet
typedef struct transport_packet {			
    // 1 byte
    unsigned char sync_byte;						//@8	bslbf
    // 2 bytes
    unsigned char transport_error_indicator;				//@1	bslbf
    unsigned char payload_unit_start_indicator;				//  1	bslbf
    unsigned char transport_priority;					//  1	bslbf
    unsigned short pid;							//  13	uimsbf
    // 1 byte
    unsigned char transport_scrambling_control;				//  2	bslbf
    unsigned char adaptation_field_control;				//  2	bslbf
    unsigned char continuity_counter;					//  4	uimsbf

    // VVV (only valid if adaptation_field_control != 1)
    // 1 byte
    unsigned char adaptation_field_length;				//  8	uimsbf
    // VVV (only valid if adaptation_field_length != 0)
    // 1 byte
    unsigned char discontinuity_indicator;				//  1	bslbf
    unsigned char random_access_indicator;				//  1	bslbf
    unsigned char elementary_stream_priority_indicator;			//  1	bslbf
    unsigned char PCR_flag;						//  1	bslbf
    unsigned char OPCR_flag;						//  1	bslbf
    unsigned char splicing_point_flag;					//  1	bslbf
    unsigned char transport_private_data_flag;				//  1	bslbf
    unsigned char adaptation_field_extension_flag;			//  1	bslbf
} transport_packet;

static char *FTType[5] = {
    "48KHz", "44.1KHz", "44.1KHz", "44.1KHz", "44.1KHz"
};

static char *AC3ModeDash[8] = {
    "1+1", "1_0", "2_0", "3_0", "2_1", "3_1", "2_2", "3_2"
};

static char *AC3Mode[8] = {
    "1+1", "1/0", "2/0", "3/0", "2/1", "3/1", "2/2", "3/2"
};

static int AC3Rate[32] = {
    32, 40, 48, 56, 64, 80, 96, 112, 128, 160,
	192, 224, 256, 320, 384, 448, 512, 576, 640,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static int VOB_ID, CELL_ID, VideoPTS, AudioPTS, AudioPTS_Start;
static unsigned char PCM_Buffer[BUFFER_SIZE];
static short *ptrPCM_Buffer = (short*)PCM_Buffer;

void Initialize_Buffer()
{
    Rdptr = Rdbfr + BUFFER_SIZE;
    Rdmax = Rdptr;

    if (SystemStream_Flag)
    {
	if (Rdptr >= Rdmax)
	    Next_Packet();
	CurrentBfr = *Rdptr++ << 24;

	if (Rdptr >= Rdmax)
	    Next_Packet();
	CurrentBfr += *Rdptr++ << 16;

	if (Rdptr >= Rdmax)
	    Next_Packet();
	CurrentBfr += *Rdptr++ << 8;

	if (Rdptr >= Rdmax)
	    Next_Packet();
	CurrentBfr += *Rdptr++;

	Fill_Next();
    }
    else
    {
	Fill_Buffer();

	CurrentBfr = (*Rdptr << 24) + (*(Rdptr+1) << 16) + (*(Rdptr+2) << 8) + *(Rdptr+3);
	Rdptr += 4;

	Fill_Next();
    }

    BitsLeft = 32;
}

// Get Video and Audio PID's from a transport stream. Uses a lookup table for A/V PIDs,
// because they are fixed to a given jap broadcaster
void GetVideoAudioPID(void)
{
    unsigned char temp_buffer[BUFFER_SIZE >> 1];
    unsigned int PIDCount[65536], max = 0;
    unsigned char sync_byte, adaptation_field_control;
    int Packet_Length, code;
    unsigned short pid;
    int i = 0;
    struct pidtable {
	int vid;
	int aid;
    } pidtable[] = {
	{ 0x0000, 0x0010 },
	{ 0x0100, 0x0110 },
	{ 0x0111, 0x0112 },
	{ 0x0140, 0x0141 },
	{ 0x0160, 0x0161 },
	{ 0x0200, 0x0210 },
	{ 0x0102, 0x0104 }, // Hitachi DT-DRX100/DT-DR1/DM-HS1, Panasonic NV-DHE10
	{ 0x1000, 0x1001 }, // Victor, Toshiba A-HD2000
	{ 0x1020, 0x1040 }, // Hitachi DT-DR20000
	{ 0x1000, 0x1011 }, // Victor, Toshiba A-HD2000
	{ 0x0102, 0x0105 }, // Hitachi DT-DRX100/DT-DR1/DM-HS1, Panasonic NV-DHE10
	{ 0xffff, 0xffff }  // Terminator
    };

    // manually search through the start of the file, for MPEG-2 transport startcodes
    for (i = 0; i + 1 < (BUFFER_SIZE >> 1); i++)
	temp_buffer[i] = Get_Byte(); // fill temp buffer

    for (i = 0; (i + (3 * 188)) < (BUFFER_SIZE >> 1); i++) {
	// search for *3* consecutive MPEG-2 transport bytes
	if (temp_buffer[i] == TS_SYNC_BYTE && temp_buffer[i + 188] == TS_SYNC_BYTE && temp_buffer[i + 2 * 188] == TS_SYNC_BYTE) {
	    SystemStream_Flag = 2;
	    dprintf("Found MPEG2TS! (stream flag %d)\n", SystemStream_Flag);
	}
    }

    // reset the file-pointer!
    File_Flag = 0;
    _lseeki64(Infile[0], 0, SEEK_SET);
    Initialize_Buffer();

    // nothing to do here, its probably plain MPEG2
    if (SystemStream_Flag != 2)
	return;

    ZeroMemory(PIDCount, sizeof(PIDCount));

    // find the PIDs
    for (i = 0; i < 100000;) {
	Packet_Length = 188;

	// locate sync byte
	do {
	    sync_byte = Get_Byte();
	} while (sync_byte != TS_SYNC_BYTE);
	Packet_Length--;

	// get pid, transport_error_indicator, payload_unit_start_indicator
	code = Get_Short();
	// decrement by 2 bytes
	Packet_Length -= 2;

	pid = (unsigned short)(code & 0x1FFF); // bits [12:0]

	// skip stream errors
	if ((code >> 15) & 0x01) {
	    SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
	    dprintf("Skipping stream error\n");
	    continue;
	}

	code = Get_Byte();
	Packet_Length--;
	adaptation_field_control = (unsigned char)((code >> 4 ) & 0x03);
	// count up this PID
	if (adaptation_field_control == 1 || adaptation_field_control == 3) {
	    PIDCount[pid]++;
	    i++;
	}

	// next packet
	SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
    }

    // find the most referenced PID (this is video)
    for (i = 0; i < 65536; i++) {
	if (max < PIDCount[i]) {
	    MPEG2_Transport_VideoPID = i;
	    max = PIDCount[i];
	}
    }

    // find the matching Audio PID from the lookup table
    for (i = 0; ; i++) {
	if (pidtable[i].vid == 0xffff)
	    break;

	if (pidtable[i].vid == MPEG2_Transport_VideoPID)
	    MPEG2_Transport_AudioPID = pidtable[i].aid;
    }

    dprintf("Using PID 0x%04x for video and 0x%04x for audio\n",
	MPEG2_Transport_VideoPID, MPEG2_Transport_AudioPID);

    // reset the file-pointer to start of file again
    File_Flag = 0;
    _lseeki64(Infile[0], 0, SEEK_SET);
    Initialize_Buffer();
}

void Next_Transport_Packet(void)
{
    int Packet_Length, Packet_Header_Length;
    int PCR_base, PCR_ext, OPCR_base, OPCR_ext;
    __int64 PCR, OPCR, PES_PTS = 0;
    unsigned int code, bytes_left;
    BOOL flag_PCR_is_here, flag_OPCR_is_here, flag_PES_PTS_is_here = FALSE;
    unsigned int AAC_Track = 0, i = 0;
    transport_packet tp ={0};		    // transport packet

    for (;;) {
	Packet_Length = 188;

	// locate sync byte
	do {
	    tp.sync_byte = Get_Byte();
	} while (tp.sync_byte != TS_SYNC_BYTE);
	Packet_Length--;

	// get pid, transport_error_indicator, payload_unit_start_indicator
	code = Get_Short();
	// decrement by 2 bytes
	Packet_Length -= 2;

	tp.pid = (unsigned short)(code & 0x1FFF); // bits [12:0]
	tp.transport_error_indicator = (unsigned char)((code >> 15) & 0x01);  // bit #15
	tp.payload_unit_start_indicator = (unsigned char)((code >> 14) & 0x01); // bit #14
	tp.transport_priority = (unsigned char)((code >> 13) & 0x01); // bit #13

	if (tp.transport_error_indicator) {
	    SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
	    continue;
	}

	code = Get_Byte();
	Packet_Length--;
	tp.transport_scrambling_control = (unsigned char)((code >> 6) & 0x03);
	tp.adaptation_field_control = (unsigned char)((code >> 4 ) & 0x03);
	tp.continuity_counter = (unsigned char)(code & 0x0F);

	// no payload
	if (tp.adaptation_field_control == 0) {
	    // || tp.adaptation_field_control == 2
	    // skip remaining bytes in current packet
	    SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
	    // abort, and circle back to top of 'for() loop'
	    continue; 
	}

	// check the Adaptation-header, only so we can determine
	// the exact #bytes to skip
	if (tp.adaptation_field_control == 2 || tp.adaptation_field_control == 3) {
	    // adaptation field is present
	    tp.adaptation_field_length = Get_Byte(); // 8-bits
	    Packet_Length--;
	    if (tp.adaptation_field_length != 0) {
		code = Get_Byte();
		Packet_Length--;

		tp.discontinuity_indicator = (unsigned char)((code >> 7) & 0x01);
		tp.random_access_indicator = (unsigned char)((code >> 6) & 0x01);
		tp.elementary_stream_priority_indicator = (unsigned char)((code >> 5) & 0x01);
		tp.PCR_flag = (unsigned char)((code >> 4) & 0x01);
		tp.OPCR_flag = (unsigned char)((code >> 3) & 0x01);
		tp.splicing_point_flag = (unsigned char)((code >> 2) & 0x01);
		tp.transport_private_data_flag = (unsigned char)((code >> 1) & 0x01);
		tp.adaptation_field_extension_flag = (unsigned char)((code >> 0) & 0x01);
		bytes_left = tp.adaptation_field_length - 1;

		// get PCR
		if (tp.PCR_flag) {
		    // we will read the next 6 bytes, to extract the program-clock-register
		    code = Get_Byte() & 0xFF; // 32..25
		    Packet_Length--;
		    PCR_base = code;

		    code = Get_Byte() & 0xFF; // 24..17
		    Packet_Length--;
		    PCR_base = (PCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 16..9
		    Packet_Length--;
		    PCR_base = (PCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 8..1
		    Packet_Length--;
		    PCR_base = (PCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 0
		    Packet_Length--;
		    PCR_base = (PCR_base << 1) | (code >> 7);

		    PCR_ext = (code & 0x80) << 1; // 8
		    code = Get_Byte() & 0xFF; // 7..0
		    Packet_Length--;

		    PCR_ext = (PCR_base << 8) | code; // 

		    PCR = PCR_base * 300 + PCR_ext;
		    flag_PCR_is_here = TRUE; // set flag

		    // we've crunched 6-bytes of the Adaptation-header.  Make a note of it.					
		    bytes_left = bytes_left - 6;
		}

		// get original PCR
		if (tp.OPCR_flag) {
		    // we will read the next 6 bytes, to extract the program-clock-register
		    code = Get_Byte() & 0xFF; // 32..25
		    Packet_Length--;
		    OPCR_base = code;

		    code = Get_Byte() & 0xFF; // 24..17
		    Packet_Length--;
		    OPCR_base = (OPCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 16..9
		    Packet_Length--;
		    OPCR_base = (OPCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 8..1
		    Packet_Length--;
		    OPCR_base = (OPCR_base << 8) | code;

		    code = Get_Byte() & 0xFF; // 0
		    Packet_Length--;
		    OPCR_base = (OPCR_base << 1) | (code >> 7);

		    OPCR_ext = (code & 0x80) << 1; // 8

		    code = Get_Byte() & 0xFF; // 7..0
		    Packet_Length--;
		    OPCR_ext = (OPCR_base << 8) | code; // 

		    OPCR = OPCR_base * 300 + OPCR_ext;
		    flag_OPCR_is_here = TRUE; // set flag

		    // we've crunched 6-bytes of the Adaptation-header.  Make a note of it.					
		    bytes_left = bytes_left - 6;
		}

		// skip the rest of adaptation field
		SKIP_TRANSPORT_PACKET_BYTES(bytes_left);
	    }
	}

	/////////////////////////////////////////////////////////////////
	// we've processed the MPEG-2 transport header, 
	// ...now, just the payload is left...

	// anything left?
	if (Packet_Length == 0)
	    continue;

	// start demux
	if (tp.pid == MPEG2_Transport_VideoPID) {
	    LOCATE;

	    // get header
	    code = Get_Short();
	    code = (code & 0xffff) << 16 | Get_Short();
	    Packet_Length -= 4;

	    // check for MPEG2-PES packet header... this contains a 'PTS'
	    if (code < 0x000001E0 || code > 0x000001E7) {
		// nope, back over
		Rdptr -= 4; 
		Packet_Length += 4;
	    } else {
		// yes, get the PTS
		Get_Short(); // MPEG2-PES total Packet_Length
		Get_Byte(); // skip a byte
		code = Get_Byte();
		Packet_Header_Length = Get_Byte();
		Packet_Length -= 5; // compensate the bytes we extracted

		// get PTS, and skip rest of PES-header
		if (code >= 0x80 && Packet_Header_Length > 4) {
		    flag_PES_PTS_is_here = TRUE;
		    code = Get_Byte();
		    PES_PTS = (code & 0x0e) << 29;
		    PES_PTS |= (Get_Short() & 0xfffe) << 14;
		    PES_PTS |= (Get_Short()>>1) & 0x7fff;
		    Packet_Length -= 5;

		    // skip the rest of the crap
		    SKIP_TRANSPORT_PACKET_BYTES(Packet_Header_Length - 5);
		} else {
		    SKIP_TRANSPORT_PACKET_BYTES(Packet_Header_Length);
		}
	    }

	    if (flag_PES_PTS_is_here)
		VideoPTS = (int)PES_PTS;

	    Rdmax = Rdptr + Packet_Length;
	    // TODO what is this? Bitrate_Meter += ((Rdmax - Rdptr) >> 1);
	    return;
	}

	if (tp.pid == MPEG2_Transport_AudioPID) {

	    // search for an MPEG-PES packet header
	    if (tp.random_access_indicator || tp.payload_unit_start_indicator) {
		LOCATE;

		code = Get_Short();
		code = (code & 0xffff) << 16 | Get_Short();
		Packet_Length -= 4;

		if (code != AUDIO_ELEMENTARY_STREAM_0 && code != AUDIO_ELEMENTARY_STREAM_1 && code != AUDIO_ELEMENTARY_STREAM_2) {
		    // nope, move back
		    Rdptr -= 4;
		    Packet_Length += 4;
		} else {
		    // yes, get the PTS
		    Get_Short();
		    Get_Byte();
		    code = Get_Byte();
		    Packet_Header_Length = Get_Byte();
		    Packet_Length -= 5;

		    if (code >= 0x80 && Packet_Header_Length > 4) {
			flag_PES_PTS_is_here = TRUE;
			code = Get_Byte();
			PES_PTS = (code & 0x0e) << 29;
			PES_PTS |= (Get_Short() & 0xfffe) << 14;
			PES_PTS |= (Get_Short()>>1) & 0x7fff;
			Packet_Length -= 5;

			SKIP_TRANSPORT_PACKET_BYTES(Packet_Header_Length - 5);
		    } else {
			SKIP_TRANSPORT_PACKET_BYTES(Packet_Header_Length);
		    }
		}
	    }

	    // done with the header, do audio data
	    if (flag_PES_PTS_is_here)
		AudioPTS = (int)PES_PTS;

	    // HACK: (timecop) is this really necessary? seems to work fine without it?!
	    LOCATE;

	    if (!aac[AAC_Track].rip && Start_Flag && !Channel[AAC_Track]) {
		// prepare to rip the AAC (open files, check if this is AAC, etc)

		// skip to the AAC sync header
		code = Get_Byte();
		code = (code & 0xff) << 8 | Get_Byte();
		Packet_Length -= 2;
		i = 0;

		while (code != 0xfff8 && Packet_Length > 0) {
		    code = (code & 0xff) << 8 | Get_Byte();
		    Packet_Length--;
		    i++;
		}

		if (code != 0xfff8) {
		    SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
		    continue;
		}

		// back up to the AAC sync word
		Packet_Length += 2;
		Rdptr -= 2;

		if (D2V_Flag || Decision_Flag) {
		    // set this here to avoid negative numbers in time display (then it's just 0's)
		    AudioPTS_Start = AudioPTS;
		    if (Method_Flag == AUDIO_DEMUXALL || Method_Flag == AUDIO_DEMUX && AAC_Track == Track_Flag) {
			Channel[AAC_Track] = FORMAT_AAC;

			// open the file
			sprintf(szBuffer, "%s (delay %dms).aac", szOutput, (AudioPTS - VideoPTS) / 90);
			aac[AAC_Track].file = fopen(szBuffer, "wb");

			DEMUX_AAC();
			aac[AAC_Track].rip = TRUE;
		    }
		}
	    } else if (aac[AAC_Track].rip) {
		DEMUX_AAC();
	    }
	}

	// nothing else to do, random data packet etc
	SKIP_TRANSPORT_PACKET_BYTES(Packet_Length);
	continue;
    }
}

void Next_Packet(void)
{
    static int i, Packet_Length, Packet_Header_Length, size;
    static unsigned int code, AUDIO_ID, VOBCELL_Count, This_Track, MPA_Track;

    // process transport stream properly
    if (SystemStream_Flag == 2) {
	Next_Transport_Packet();
	return;
    }

    for (;;) {
	code = Get_Short();
	code = (code<<16) + Get_Short();

	while ((code & 0xffffff00) != 0x00000100)
	    code = (code<<8) + Get_Byte();

	switch (code)
	{
	case PACK_START_CODE:
	    Rdptr += 8;
	    VOBCELL_Count = 0;
	    break;

	case PRIVATE_STREAM_2:
	    Packet_Length = Get_Short();

	    if (++VOBCELL_Count==2)
	    {
		Rdptr += 25;
		VOB_ID = Get_Short();
		CELL_ID = Get_Short();
		Rdptr += Packet_Length - 29;

		sprintf(szBuffer, "%d", VOB_ID);
		SetDlgItemText(hDlg, IDC_VOB_ID, szBuffer);

		sprintf(szBuffer, "%d", CELL_ID);
		SetDlgItemText(hDlg, IDC_CELL_ID, szBuffer);
	    }
	    else
		Rdptr += Packet_Length;
	    break;

	case PRIVATE_STREAM_1:
	    Packet_Length = Get_Short();

	    Rdptr ++;	// +1
	    code = Get_Byte();	// +1
	    Packet_Header_Length = Get_Byte();	// +1

	    if (code>=0x80)
	    {
		code = Get_Byte();
		AudioPTS = (code & 0x0e) << 29;

		AudioPTS |= (Get_Short() & 0xfffe) << 14;
		AudioPTS |= (Get_Short()>>1) & 0x7fff;

		Rdptr += Packet_Header_Length-5;
	    }
	    else
		Rdptr += Packet_Header_Length;

	    AUDIO_ID = Get_Byte();	// +1
	    Packet_Length -= Packet_Header_Length+4;

	    if (AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL)
	    {
		Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_AC3;

		LOCATE;

		if (!ac3[This_Track].rip && Start_Flag && !Channel[This_Track])
		{
		    Channel[This_Track] = FORMAT_AC3;

		    code = Get_Byte();
		    code = (code & 0xff)<<8 | Get_Byte();
		    i = 0;

		    while (code!=0xb77)
		    {
			code = (code & 0xff)<<8 | Get_Byte();
			i++;
		    }

		    Get_Short();
		    ac3[This_Track].rate = (Get_Byte()>>1) & 0x1f;
		    Get_Byte();
		    ac3[This_Track].mode = (Get_Byte()>>5) & 0x07;

		    Rdptr -= 7; Packet_Length -= i;

		    if (D2V_Flag || Decision_Flag)
		    {
			if (Decision_Flag && This_Track==Track_Flag)
			{
			    InitialAC3();

			    DECODE_AC3

				ac3[This_Track].rip = true;
			}
			else if (Method_Flag==AUDIO_DECODE && This_Track==Track_Flag)
			{
			    sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps %s.wav", szOutput, This_Track+1, 
				AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], FTType[SRC_Flag]);

			    strcpy(pcm.filename, szBuffer);
			    pcm.file = fopen(szBuffer, "wb");

			    StartWAV(pcm.file);
			    pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;

			    if (SRC_Flag)
			    {
				DownWAV(pcm.file);
				InitialSRC();
			    }

			    if (pcm.delay > 0)
			    {
				if (SRC_Flag)
				    pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

				for (i=0; i<pcm.delay; i++)
				    fputc(0, pcm.file);

				pcm.size += pcm.delay;
				pcm.delay = 0;
			    }

			    InitialAC3();

			    DECODE_AC3

				if (-pcm.delay > size)
				    pcm.delay += size;
				else
				{
				    if (SRC_Flag)
					Wavefs44(pcm.file, size+pcm.delay, AC3Dec_Buffer-pcm.delay);
				    else
					fwrite(AC3Dec_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

				    pcm.size += size+pcm.delay;
				    pcm.delay = 0;
				}

				ac3[This_Track].rip = true;
			}
			else if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && This_Track==Track_Flag)
			{
			    sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps DELAY %dms.ac3", szOutput, This_Track+1, 
				AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], (AudioPTS-VideoPTS)/90);

			    ac3[This_Track].file = fopen(szBuffer, "wb");

			    DEMUX_AC3

				ac3[This_Track].rip = true;
			}
		    }
		}
		else if (ac3[This_Track].rip)
		{
		    if (Decision_Flag)
			DECODE_AC3
		    else if (Method_Flag==AUDIO_DECODE)
		    {
			DECODE_AC3

			    if (-pcm.delay > size)
				pcm.delay += size;
			    else
			    {
				if (SRC_Flag)
				    Wavefs44(pcm.file, size+pcm.delay, AC3Dec_Buffer-pcm.delay);
				else
				    fwrite(AC3Dec_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

				pcm.size += size+pcm.delay;
				pcm.delay = 0;
			    }
		    }
		    else
			DEMUX_AC3
		}
	    }
	    else if (AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL)
	    {
		Rdptr += 6; Packet_Length -= 6; This_Track = AUDIO_ID-SUB_PCM;

		LOCATE;

		if (!pcm.rip && Start_Flag && !Channel[This_Track])
		{
		    Channel[This_Track] = FORMAT_LPCM;

		    if (D2V_Flag)
		    {
			if (Method_Flag==AUDIO_DECODE && This_Track==Track_Flag)
			{
			    pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;

			    sprintf(szBuffer, "%s LPCM T%02d %s.wav", szOutput, This_Track+1, FTType[SRC_Flag]);
			    strcpy(pcm.filename, szBuffer);

			    pcm.file = fopen(szBuffer, "wb");
			    StartWAV(pcm.file);

			    if (SRC_Flag)
			    {
				DownWAV(pcm.file);
				InitialSRC();
			    }

			    if (pcm.delay > 0)
			    {
				if (SRC_Flag)
				    pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

				for (i=0; i<pcm.delay; i++)
				    fputc(0, pcm.file);

				pcm.size += pcm.delay;
				pcm.delay = 0;
			    }

			    if (-pcm.delay > Packet_Length)
				pcm.delay += Packet_Length;
			    else
			    {
				DEMUX_PCM

				    if (SRC_Flag)
					Wavefs44(pcm.file, size+pcm.delay, PCM_Buffer-pcm.delay);
				    else
					fwrite(PCM_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

				pcm.size += size+pcm.delay;
				pcm.delay = 0;
			    }

			    pcm.rip = true;
			}
		    }
		}
		else if (pcm.rip)
		{
		    if (-pcm.delay > Packet_Length)
			pcm.delay += Packet_Length;
		    else
		    {
			DEMUX_PCM

			    if (SRC_Flag)
				Wavefs44(pcm.file, size+pcm.delay, PCM_Buffer-pcm.delay);
			    else
			    {
				fwrite(PCM_Buffer-pcm.delay, size+pcm.delay, 1, pcm.file);

				if (Norm_Flag)
				    for (i=0; i<(size>>1); i++)
					if (Sound_Max < abs(ptrPCM_Buffer[i]))
					    Sound_Max = abs(ptrPCM_Buffer[i]);
			    }

			    pcm.size += size+pcm.delay;
			    pcm.delay = 0;
		    }
		}
	    }
	    else if (AUDIO_ID>=SUB_DTS && AUDIO_ID<SUB_DTS+CHANNEL)
	    {
		Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_DTS;

		LOCATE;

		if (!dts[This_Track].rip && Start_Flag && !Channel[This_Track])
		{
		    Channel[This_Track] = FORMAT_DTS;

		    if (D2V_Flag)
		    {
			if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && This_Track==Track_Flag)
			{
			    sprintf(szBuffer, "%s DTS T%02d DELAY %dms.dts", szOutput, This_Track+1, (AudioPTS-VideoPTS)/90);

			    dts[This_Track].file = fopen(szBuffer, "wb");

			    DEMUX_DTS

				dts[This_Track].rip = true;
			}
		    }
		}
		else if (dts[This_Track].rip)
		    DEMUX_DTS
	    }
	    Rdptr += Packet_Length;
	    break;

	case AUDIO_ELEMENTARY_STREAM_7:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_6:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_5:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_4:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_3:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_2:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_1:
	    MPA_Track++;
	case AUDIO_ELEMENTARY_STREAM_0:
	    Packet_Length = Get_Short()-1;
	    code = Get_Byte();

	    if ((code & 0xc0)==0x80)
	    {
		code = Get_Byte();	// +1
		Packet_Header_Length = Get_Byte();	// +1

		if (code>=0x80)
		{
		    code = Get_Byte();

		    AudioPTS = (code & 0x0e) << 29;
		    AudioPTS |= (Get_Short() & 0xfffe) << 14;
		    AudioPTS |= (Get_Short()>>1) & 0x7fff;

		    Rdptr += Packet_Header_Length-5;
		}
		else
		    Rdptr += Packet_Header_Length;

		Packet_Length -= Packet_Header_Length+2;

		LOCATE;

		if (!mpa[MPA_Track].rip && Start_Flag && !Channel[MPA_Track])
		{
		    Channel[MPA_Track] = FORMAT_MPA;

		    code = Get_Byte();
		    code = (code & 0xff)<<8 | Get_Byte();
		    i = 0;

		    while (code<0xfff0)
		    {
			code = (code & 0xff)<<8 | Get_Byte();
			i++;
		    }

		    Rdptr -= 2; Packet_Length -= i;

		    if (D2V_Flag)
		    {
			if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && MPA_Track==Track_Flag)
			{
			    sprintf(szBuffer, "%s MPA T%02d DELAY %dms.mpa", szOutput, MPA_Track+1, (AudioPTS-VideoPTS)/90);
			    mpa[MPA_Track].file = fopen(szBuffer, "wb");

			    DEMUX_MPA

				mpa[MPA_Track].rip = true;
			}
		    }
		}
		else if (mpa[MPA_Track].rip)
		    DEMUX_MPA
	    }

	    Rdptr += Packet_Length;
	    MPA_Track = 0;
	    break;

	case VIDEO_ELEMENTARY_STREAM:
	    Packet_Length = Get_Short();
	    Rdmax = Rdptr + Packet_Length;

	    code = Get_Byte();

	    if ((code & 0xc0)==0x80)
	    {
		code = Get_Byte();
		Packet_Header_Length = Get_Byte();

		if (code>=0x80 && !Start_Flag)
		{
		    code = Get_Byte();
		    VideoPTS = (code & 0x0e) << 29;
		    VideoPTS |= (Get_Short() & 0xfffe) << 14;
		    VideoPTS |= (Get_Short()>>1) & 0x7fff;

		    Rdptr += Packet_Header_Length-5;
		}
		else
		    Rdptr += Packet_Header_Length;

		Bitrate_Monitor += Rdmax-Rdptr;
		return;
	    }
	    else
		Rdptr += Packet_Length-1;
	    break;

	default:
	    if (code>=SYSTEM_START_CODE)
	    {
		Packet_Length = Get_Short();
		Rdptr += Packet_Length;
	    }
	    break;
	}
    }
}

unsigned int Get_Bits_All(unsigned int N)
{
    N -= BitsLeft;
    Val = (CurrentBfr << (32 - BitsLeft)) >> (32 - BitsLeft);

    if (N)
	Val = (Val << N) + (NextBfr >> (32 - N));

    CurrentBfr = NextBfr;
    BitsLeft = 32 - N;
    Fill_Next();

    return Val;
}

void Flush_Buffer_All(unsigned int N)
{
    CurrentBfr = NextBfr;
    BitsLeft = BitsLeft + 32 - N;
    Fill_Next();
}

void Fill_Buffer()
{
    Read = _read(Infile[File_Flag], Rdbfr, BUFFER_SIZE);

    if (Read < BUFFER_SIZE)
	Next_File();

    if (KeyOp_Flag && (Rdbfr[20] & 0x10))
    {
	BufferOp(Rdbfr, lfsr0, lfsr1);
	Rdbfr[20] &= ~0x10;
    }

    Rdptr = Rdbfr;

    if (SystemStream_Flag)
	Rdmax -= BUFFER_SIZE;
    else
	Bitrate_Monitor += Read;
}

void Next_File()
{
    int i;

    if (File_Flag < File_Limit-1)
    {
	File_Flag++;

	process.run = 0;
	for (i=0; i<File_Flag; i++)
	    process.run += process.length[i];

	_lseeki64(Infile[File_Flag], 0, SEEK_SET);
	_read(Infile[File_Flag], Rdbfr + Read, BUFFER_SIZE - Read);
    }
    else
    {
	Fault_Flag = NO_MORE_FILES_ERROR;
	dprintf("Cannot read anymore of this file!\n");
    }
}

static char *AspectRatio[5] = {
    "", "1 : 1", "4 : 3", "16 : 9", "2.21 : 1"
};

void UpdateInfo()
{
    static int Old_NTSC_Purity;
    int pts;

    sprintf(szBuffer, "%s", AspectRatio[aspect_ratio_information]);
    SetDlgItemText(hDlg, IDC_ASPECT_RATIO, szBuffer);

    sprintf(szBuffer, "%.3f fps", Frame_Rate);
    SetDlgItemText(hDlg, IDC_FRAME_RATE, szBuffer);

    switch (Channel[Track_Flag])
    {
    case FORMAT_AC3:
	sprintf(szBuffer, "DD %s %d", AC3Mode[ac3[Track_Flag].mode], AC3Rate[ac3[Track_Flag].rate]);
	SetDlgItemText(hDlg, IDC_AUDIO_TYPE, szBuffer);
	break;

    case FORMAT_MPA:
	SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "MPEG Audio");
	break;

    case FORMAT_LPCM:
	SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "Linear PCM");
	break;

    case FORMAT_DTS:
	SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "DTS");
	break;

    default:
	SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "");
	break;
    }

    if (SystemStream_Flag && process.locate!=LOCATE_INIT)
    {
	// TODO use Video PTS for this
	pts = (AudioPTS - AudioPTS_Start) / 90000;
	sprintf(szBuffer, "%d:%02d:%02d", pts/3600, (pts%3600)/60, pts%60);
	SetDlgItemText(hDlg, IDC_TIMESTAMP, szBuffer);
    }
    else
	SetDlgItemText(hDlg, IDC_TIMESTAMP, "");

    if (process.locate==LOCATE_RIP)
    {
	if (NTSC_Purity || FILM_Purity)
	{
	    if (!FILM_Purity)
	    {
		if (frame_rate==25)
		    sprintf(szBuffer, "PAL");
		else
		    sprintf(szBuffer, "NTSC");
	    }
	    else if (!NTSC_Purity)
		sprintf(szBuffer, "FILM");
	    else if (NTSC_Purity > Old_NTSC_Purity)
		sprintf(szBuffer, "NTSC %2d %%", 100 - (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));
	    else
		sprintf(szBuffer, "FILM %2d %%", (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));

	    Old_NTSC_Purity = NTSC_Purity;
	    SetDlgItemText(hDlg, IDC_VIDEO_TYPE, szBuffer);
	}

	if (process.op)
	{
	    process.ed = timeGetTime();
	    process.elapsed = (process.ed-process.op)/1000;
	    process.percent = (float)(100.0*(process.run-process.start+_telli64(Infile[File_Flag]))/(process.end-process.start));
	    process.remain = (int)((process.ed-process.op)*(100.0-process.percent)/process.percent)/1000;

	    sprintf(szBuffer, "%d:%02d:%02d", process.elapsed/3600, (process.elapsed%3600)/60, process.elapsed%60);
	    SetDlgItemText(hDlg, IDC_ELAPSED, szBuffer);

	    sprintf(szBuffer, "%d:%02d:%02d", process.remain/3600, (process.remain%3600)/60, process.remain%60);
	    SetDlgItemText(hDlg, IDC_REMAIN, szBuffer);
	}
    }
    else if (GetDlgItemText(hDlg, IDC_ELAPSED, szBuffer, 9))
    {
	SetDlgItemText(hDlg, IDC_VIDEO_TYPE, "");
	SetDlgItemText(hDlg, IDC_FRAME_TYPE, "");
	SetDlgItemText(hDlg, IDC_CODED_NUMBER, "");
	SetDlgItemText(hDlg, IDC_PLAYBACK_NUMBER, "");
	SendDlgItemMessage(hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
	SetDlgItemText(hDlg, IDC_FILE, "");
	SetDlgItemText(hDlg, IDC_FILE_SIZE, "");
	SetDlgItemText(hDlg, IDC_ELAPSED, "");
	SetDlgItemText(hDlg, IDC_REMAIN, "");
	SetDlgItemText(hDlg, IDC_FPS, "");
    }
}
