/* 
 *	Copyright (C) Chia-chen Kuo - April 2001
 *
 *  This file is part of DVD2AVI, a free MPEG-2 decoder
 *	
 *  DVD2AVI is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVI is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "filter.h"

#define MAX_AVI_SIZE	2073600000

__forceinline static void Store_RGB24(unsigned char *src[], DWORD frame);
__forceinline static void Store_YUY2(unsigned char *src[], DWORD frame);

static void FlushRGB24(void);
static void FlushYUY2(void);
static void AVIKill(void);

static PAVIFILE pfile;
static PAVISTREAM ps, psCompressed;
static AVICOMPRESSOPTIONS opts;
static AVISTREAMINFO strinfo;
static ICCOMPRESSFRAMES iccf;
static COMPVARS compvars;
static BITMAPINFOHEADER birgb, biyuv;

static bool AVI_Init;
static char VideoOut[_MAX_PATH];
static unsigned char *y444;
static int avi_size, avi_count, frame_count;
static int playback, Old_Playback;
static long frame_size;
static bool TFF, RFF, TFB, BFB, frame_type;

static bool Resize_Flag;
static int Resize_Width, Resize_Height;

static char *FrameType[] = {
	"Interlaced", "Progressive"
};

void Write_Frame(unsigned char *src[], D2VData d2v, DWORD frame)
{
	int repeat;

	if (Fault_Flag) {
		if (Fault_Flag < CRITICAL_ERROR_LEVEL)
		{
			SetDlgItemText(hDlg, IDC_INFO, "V.E.!");
			Fault_Flag = 0;		// fault tolerance
		}
		else
		{
			if (AVI_Flag)
				AVIKill();
			ThreadKill();
		}
	}

	frame_type = d2v.pf;
	TFF = d2v.trf>>1;
	RFF = d2v.trf & 0x01;

	if (!frame)
	{
		char *ext, szTemp[_MAX_PATH];
		int i;

		AVI_Init = true; TFB = BFB = false; Old_Playback = 0;
		frame_size = 0; avi_size = 0; avi_count = 1; frame_count = 0; playback = 0;

		Resize_Flag = false;
		Clip_Width = Resize_Width = Coded_Picture_Width;
		Clip_Height = Resize_Height = Coded_Picture_Height;
		CLIP_AREA = HALF_CLIP_AREA = CLIP_STEP = CLIP_HALF_STEP = 0;

		if (Luminance_Flag)
			InitializeFilter();

		if (ClipResize_Flag)
		{
			if (Clip_Top || Clip_Bottom || Clip_Left || Clip_Right)
			{
				Clip_Width -= Clip_Left+Clip_Right;
				Clip_Height -= Clip_Top+Clip_Bottom;
				Resize_Width = Clip_Width;
				Resize_Height = Clip_Height;

				CLIP_AREA = Coded_Picture_Width * Clip_Top;
				HALF_CLIP_AREA = (Coded_Picture_Width>>1) * Clip_Top;

				CLIP_STEP = Coded_Picture_Width * Clip_Top + Clip_Left;
				CLIP_HALF_STEP = (Coded_Picture_Width>>1) * Clip_Top + (Clip_Left>>1);
			}

			if (Squeeze_Width || Squeeze_Height)
			{
				Resize_Flag = true;
				Resize_Width -= Squeeze_Width;
				Resize_Height -= Squeeze_Height;
			}
		}

		NINE_CLIP_WIDTH = Clip_Width * 9;
		QUAD_CLIP_WIDTH = Clip_Width<<2;
		DOUBLE_CLIP_WIDTH = Clip_Width<<1;
		HALF_CLIP_WIDTH = Clip_Width>>1;

		LUM_AREA = Coded_Picture_Width * Clip_Height;
		DOUBLE_WIDTH = Coded_Picture_Width<<1;
		HALF_WIDTH = Coded_Picture_Width>>1;
		HALF_WIDTH_D8 = (Coded_Picture_Width>>1) - 8;
		PROGRESSIVE_HEIGHT = (Coded_Picture_Height>>1) - 2;
		INTERLACED_HEIGHT = (Coded_Picture_Height>>2) - 2;
		RGB_DOWN1 = Clip_Width * (Clip_Height - 1) * 3;
		RGB_DOWN2 = Clip_Width * (Clip_Height - 2) * 3;

		sprintf(szBuffer, "DVD2AVI - [%d / %d] ", File_Flag+1, File_Limit);
		ext = strrchr(Infilename[File_Flag], '\\');
		strncat(szBuffer, ext+1, strlen(Infilename[0])-(int)(ext-Infilename[0]));

		sprintf(szTemp, " %d x %d", Resize_Width, Resize_Height);
		strcat(szBuffer, szTemp);

		if (Resize_Width%32 == 0)
			strcat(szBuffer, " [32X]");
		else if (Resize_Width%16 == 0)
			strcat(szBuffer, " [16X]");
		else if (Resize_Width%8 == 0)
			strcat(szBuffer, " [8X]");
		else
			strcat(szBuffer, " [ ]");

		if (Resize_Height%32 == 0)
			strcat(szBuffer, "[32X]");
		else if (Resize_Height%16 == 0)
			strcat(szBuffer, "[16X]");
		else if (Resize_Height%8 == 0)
			strcat(szBuffer, "[8X]");
		else
			strcat(szBuffer, "[ ]");

		// append A/V PID info
		if (SystemStream_Flag == 2) {
		    sprintf(szTemp, " V:0x%04x A:0x%04x", MPEG2_Transport_VideoPID, MPEG2_Transport_AudioPID);
		    strcat(szBuffer, szTemp);
		}

		SetWindowText(hWnd, szBuffer);

		ResizeWindow(Clip_Width, Clip_Height);

		ZeroMemory(&birgb, sizeof(BITMAPINFOHEADER));
		birgb.biSize = sizeof(BITMAPINFOHEADER);
		birgb.biWidth = Clip_Width;
		birgb.biHeight = Clip_Height;
		birgb.biPlanes = 1;
		birgb.biBitCount = 24;
		birgb.biCompression = BI_RGB;
		birgb.biSizeImage = Clip_Width * Clip_Height * 3;

		ZeroMemory(&biyuv, sizeof(BITMAPINFOHEADER));
		biyuv = birgb;
		biyuv.biBitCount = 16;
		biyuv.biCompression = mmioFOURCC('Y','U','Y','2');
		biyuv.biSizeImage = Clip_Width * Clip_Height * 2;

		if (process.locate==LOCATE_RIP && FO_Flag==FO_SWAP)
		{
			if (TFF)
				BFB = true;
			else
				TFB = true;

			// mask
			ZeroMemory(rgb24, Coded_Picture_Width * Clip_Height * 3);
			for (i=0; i<Coded_Picture_Width * Coded_Picture_Height * 2; i+=2)
			{
				yuy2[i] = 0;
				yuy2[i+1] = 128;
			}
		}

		if (FO_Flag!=FO_FILM)
		{
			if (TFF && FO_Flag!=FO_SWAP  ||  !TFF && FO_Flag==FO_SWAP)
				SetDlgItemText(hDlg, IDC_INFO, "T");
			else
				SetDlgItemText(hDlg, IDC_INFO, "B");
		}
	}

	repeat = DetectVideoType(frame, d2v.trf);

	if (FO_Flag!=FO_FILM || repeat)
	{
		if (Store_Flag==STORE_YUY2)
			Store_YUY2(src, frame);
		else
			Store_RGB24(src, frame);
	}

	if (FO_Flag==FO_FILM && repeat==2)
	{
		if (Store_Flag==STORE_YUY2)
			Store_YUY2(src, frame);
		else
			Store_RGB24(src, frame);
	}

	if (Info_Flag && process.locate==LOCATE_RIP)
	{
		sprintf(szBuffer, "%s", FrameType[frame_type]);
		SetDlgItemText(hDlg, IDC_FRAME_TYPE, szBuffer);

		sprintf(szBuffer, "%d", frame+1);
		SetDlgItemText(hDlg, IDC_CODED_NUMBER, szBuffer);

		sprintf(szBuffer, "%d", playback);
		SetDlgItemText(hDlg, IDC_PLAYBACK_NUMBER, szBuffer);

		if (AVI_Flag)
		{
			sprintf(szBuffer, "%d", avi_count-1);
			SetDlgItemText(hDlg, IDC_FILE, szBuffer);
	
			sprintf(szBuffer, "%d MB", avi_size>>20);
			SetDlgItemText(hDlg, IDC_FILE_SIZE, szBuffer);
		}

		if ((frame & 63) == 63)
		{
			process.ed = timeGetTime();

			sprintf(szBuffer, "%.1f", 1000.0*(playback-Old_Playback)/(process.ed-process.mi+1));
			SetDlgItemText(hDlg, IDC_FPS, szBuffer);
			SendDlgItemMessage(hDlg, IDC_PROGRESS, PBM_SETPOS, (int)(Bitrate_Monitor*Frame_Rate/(playback-Old_Playback))>>7, 0);

			Bitrate_Monitor = 0;
			process.mi = process.ed;
			Old_Playback = playback;
		}
	}

	if (AVI_Flag && avi_size>=MAX_AVI_SIZE)
		AVIKill();
}

static void Store_RGB24(unsigned char *src[], DWORD frame)
{
	if (AVI_Flag && AVI_Init)
	{
		if (!frame)
		{
			compvars.cbSize = sizeof(compvars);

			if (!ICCompressorChoose(hWnd, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME,
				&birgb, NULL, &compvars, NULL))
				ThreadKill();

			// set AVI header
			ZeroMemory(&strinfo, sizeof(AVISTREAMINFO));
			strinfo.fccType					= streamtypeVIDEO;
			strinfo.fccHandler				= compvars.fccHandler;
			strinfo.dwQuality				= -1;
			strinfo.dwSuggestedBufferSize	= birgb.biSizeImage;
			strinfo.dwScale	= 1000;
			strinfo.dwRate = (unsigned int)((FO_Flag==FO_FILM) ? frame_rate*800 : frame_rate*1000);

			// set AVI save options
			opts.cbFormat = sizeof(birgb);
			opts.fccType = streamtypeVIDEO;
			opts.fccHandler = compvars.fccHandler;
			opts.dwKeyFrameEvery = compvars.lKey;
			opts.dwQuality = compvars.lQ;
			opts.dwBytesPerSecond = compvars.lDataRate<<10;
			opts.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES | AVICOMPRESSF_DATARATE;
			opts.dwInterleaveEvery = 0;
			opts.lpFormat = &birgb;
			opts.cbParms = compvars.cbState;
			opts.lpParms = compvars.lpState;
		}

		AVI_Init = false; avi_size = 0;
		AVIFileInit();

		sprintf(VideoOut, "%s.%02d.avi", szOutput, avi_count++);

		if (AVIFileOpen(&pfile, VideoOut, OF_WRITE | OF_CREATE, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIFileCreateStream(pfile, &ps, &strinfo) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIStreamSetFormat(psCompressed, 0, &birgb, birgb.biSize) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}
	}

	if (chroma_format==CHROMA420)
	{
		conv420to422(src[1], u422, frame_type);
		conv420to422(src[2], v422, frame_type);

		conv422to444(u422, u444);
		conv422to444(v422, v444);
	}
	else
	{
		conv422to444(src[1], u444);
		conv422to444(src[2], v444);
	}

	if (Luminance_Flag)
	{
		LuminanceFilter(src[0], lum);
		y444 = lum;
	}
	else
		y444 = src[0];

	if (BFB)
	{
		conv444toRGB24odd(y444, u444, v444, rgb24);
		TFB = true;
		FlushRGB24();

		conv444toRGB24even(y444, u444, v444, rgb24);
		BFB = true;
		FlushRGB24();
	}
	else
	{
		conv444toRGB24even(y444, u444, v444, rgb24);
		BFB = true;
		FlushRGB24();

		conv444toRGB24odd(y444, u444, v444, rgb24);
		TFB = true;
		FlushRGB24();
	}

	if (FO_Flag!=FO_FILM && RFF)
		if (TFF)
		{
			TFB = true;
			FlushRGB24();
		}
		else
		{
			BFB = true;
			FlushRGB24();
		}
}

static void FlushRGB24()
{
	if (TFB & BFB)
	{
		if (AVI_Flag)
		{
			if (AVIStreamWrite(psCompressed, frame_count++, 1, rgb24,
				birgb.biSizeImage, 0, NULL, &frame_size) != AVIERR_OK)
			{
				AVIKill();
				ThreadKill();
			}

			avi_size += frame_size;
		}

		ShowFrame(false);

		playback++;
		TFB = BFB = false;
	}
}

static void Store_YUY2(unsigned char *src[], DWORD frame)
{
	if (AVI_Flag && AVI_Init)
	{
		if (!frame)
		{
			compvars.cbSize = sizeof(compvars);

			if (!ICCompressorChoose(hWnd, ICMF_CHOOSE_DATARATE | ICMF_CHOOSE_KEYFRAME,
				&biyuv, NULL, &compvars, NULL))
				ThreadKill();

			// set AVI header
			ZeroMemory(&strinfo, sizeof(AVISTREAMINFO));
			strinfo.fccType					= streamtypeVIDEO;
			strinfo.fccHandler				= compvars.fccHandler;
			strinfo.dwQuality				= -1;
			strinfo.dwSuggestedBufferSize	= biyuv.biSizeImage;
			strinfo.dwScale	= 1000;
			strinfo.dwRate = (unsigned int)((FO_Flag==FO_FILM) ? frame_rate*800 : frame_rate*1000);

			// set AVI save options
			opts.cbFormat = sizeof(biyuv);
			opts.fccType = streamtypeVIDEO;
			opts.fccHandler = compvars.fccHandler;
			opts.dwKeyFrameEvery = compvars.lKey;
			opts.dwQuality = compvars.lQ;
			opts.dwBytesPerSecond = compvars.lDataRate<<10;
			opts.dwFlags = AVICOMPRESSF_VALID | AVICOMPRESSF_KEYFRAMES | AVICOMPRESSF_DATARATE;
			opts.dwInterleaveEvery = 0;
			opts.lpFormat = &biyuv;
			opts.cbParms = compvars.cbState;
			opts.lpParms = compvars.lpState;

			iccf.dwRate = strinfo.dwRate;
			iccf.dwScale = strinfo.dwScale;
			iccf.lQuality = compvars.lQ;
			iccf.lDataRate = compvars.lDataRate<<10;
			iccf.lKeyRate = compvars.lKey;

			ICSendMessage(compvars.hic, ICM_COMPRESS_FRAMES_INFO, (WPARAM)&iccf, (DWORD)sizeof(ICCOMPRESSFRAMES));
		}

		AVI_Init = false; avi_size = 0;
		AVIFileInit();

		sprintf(VideoOut, "%s.%02d.avi", szOutput, avi_count++);

		if (AVIFileOpen(&pfile, VideoOut, OF_WRITE | OF_CREATE, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIFileCreateStream(pfile, &ps, &strinfo) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (AVIStreamSetFormat(psCompressed, 0, &birgb, birgb.biSize) != AVIERR_OK)
		{
			AVIKill();
			ThreadKill();
		}

		if (!frame && !ICSeqCompressFrameStart(&compvars, (LPBITMAPINFO)&biyuv))
		{
			AVIKill();
			ThreadKill();
		}
	}

	if (chroma_format==CHROMA420)
	{
		conv420to422(src[1], u422, frame_type);
		conv420to422(src[2], v422, frame_type);
	}
	else
	{
		u422 = src[1];
		v422 = src[2];
	}

	if (Luminance_Flag)
	{
		LuminanceFilter(src[0], lum);
		y444 = lum;
	}
	else
		y444 = src[0];

	if (BFB)
	{
		conv422toyuy2odd(y444, u422, v422, yuy2);
		TFB = true;
		FlushYUY2();

		conv422toyuy2even(y444, u422, v422, yuy2);
		BFB = true;
		FlushYUY2();
	}
	else
	{
		conv422toyuy2even(y444, u422, v422, yuy2);
		BFB = true;
		FlushYUY2();

		conv422toyuy2odd(y444, u422, v422, yuy2);
		TFB = true;
		FlushYUY2();
	}

	if (FO_Flag!=FO_FILM && RFF)
		if (TFF)
		{
			TFB = true;
			FlushYUY2();
		}
		else
		{
			BFB = true;
			FlushYUY2();
		}
}

static void FlushYUY2()
{
	void *yuy2c;
	int key_flag;
	long yuy2c_size;

	if (TFB & BFB)
	{
		if (AVI_Flag)
		{
			if ((yuy2c = ICSeqCompressFrame(&compvars, 0, yuy2, &key_flag, &yuy2c_size)) == NULL)
			{
				AVIKill();
				ThreadKill();
			}

			if (AVIStreamWrite(ps, frame_count++, 1, yuy2c, yuy2c_size,
				key_flag ? AVIIF_KEYFRAME : 0, NULL, &frame_size) != AVIERR_OK)
			{
				AVIKill();
				ThreadKill();
			}

			avi_size += frame_size;
		}

		ShowFrame(false);

		playback++;
		TFB = BFB = false;
	}
}

int DetectVideoType(int frame, int trf)
{
	static int Old_TRF, Repeat_On, Repeat_Off;
	static bool Repeat_Init;

	if (frame)
	{
		if ((trf & 3) == ((Old_TRF+1) & 3))
			FILM_Purity++;
		else
			NTSC_Purity++;
	}
	else
	{
		FILM_Purity = NTSC_Purity = Repeat_On = Repeat_Off = 0;
		Repeat_Init = false;
	}

	Old_TRF = trf;

	if (trf & 1)
		Repeat_On ++;
	else
		Repeat_Off ++;

	if (Repeat_Init)
	{
		if (Repeat_Off-Repeat_On == 5)
		{
			Repeat_Off = Repeat_On = 0;
			return 0;
		}
		else if (Repeat_On-Repeat_Off == 5)
		{
			Repeat_Off = Repeat_On = 0;
			return 2;
		}
	}
	else
	{
		if (Repeat_Off-Repeat_On == 3)
		{
			Repeat_Off = Repeat_On = 0;
			Repeat_Init = true;
			return 0;
		}
		else if (Repeat_On-Repeat_Off == 3)
		{
			Repeat_Off = Repeat_On = 0;
			Repeat_Init = true;
			return 2;
		}
	}

	return 1;
}

static void AVIKill()
{
	AVI_Init = true; frame_count = 0;

	if (ps)
		AVIStreamClose(ps);

	if (psCompressed)
		AVIStreamClose(psCompressed);

	if (pfile)
		AVIFileClose(pfile);

	if (Stop_Flag || Fault_Flag)
	{
		AVIFileExit();

		if (Store_Flag!=STORE_RGB24)
			ICSeqCompressFrameEnd(&compvars);

		ICCompressorFree(&compvars);
	}
}

void ShowFrame(bool move)
{
	if (!Check_Flag)
		return;

	if (Store_Flag==STORE_YUY2)
	{
		if ((Display_Flag || move) && DDOverlay_Flag)
		{
			POINT point = {0, 0};
			RECT orect, prect;

			if (!move && IDirectDrawSurface_Lock(lpOverlay, NULL, &ddsd, 0, NULL)==DD_OK)
			{
				int i;
				unsigned char *src = yuy2;		
				unsigned char *dst = (unsigned char *)ddsd.lpSurface;

				for (i=0; i<Clip_Height; i++)
				{
					memcpy(dst, src, DOUBLE_CLIP_WIDTH);
					src += DOUBLE_CLIP_WIDTH;
					dst += ddsd.lPitch;
				}
		
				IDirectDrawSurface_Unlock(lpOverlay, NULL);
			}

			SetRect(&orect, 0, 0, Coded_Picture_Width, Coded_Picture_Height);

			ClientToScreen(hWnd, &point);
			prect.left = point.x;
			prect.right = point.x + Coded_Picture_Width;
			prect.top = point.y;
			prect.bottom = point.y + Coded_Picture_Height;

			if (prect.left < 0)
			{
				orect.left = -prect.left;
				prect.left = 0;
			}

			if (prect.top < 0)
			{
				orect.top = -prect.top;
				prect.top = 0;
			}

			if (prect.right > GetSystemMetrics(SM_CXSCREEN))
			{
				orect.right = Coded_Picture_Width + GetSystemMetrics(SM_CXSCREEN) - prect.right;
				prect.right = GetSystemMetrics(SM_CXSCREEN);
			}

			if (prect.bottom > GetSystemMetrics(SM_CYSCREEN))
			{
				orect.bottom = Coded_Picture_Height + GetSystemMetrics(SM_CYSCREEN) - prect.bottom;
				prect.bottom = GetSystemMetrics(SM_CYSCREEN);
			}

			IDirectDrawSurface_UpdateOverlay(lpOverlay, &orect, lpPrimary, &prect,
				DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE, &ddofx);

//			IDirectDrawSurface_SetOverlayPosition(lpOverlay, prect.left, prect.top);
		}
	}
	else
	{
		if (Display_Flag || move)
			SetDIBitsToDevice(hDC, 0, 0, Clip_Width, Clip_Height,
				0, 0, 0, Clip_Height, rgb24, (LPBITMAPINFO)&birgb, DIB_RGB_COLORS);
	}
}
